#!/usr/bin/env ruby

#  1750, Sat 31 Dec 11 (NZDT)  v2.1.0
#  2015, Tue  1 Jun 10 (NZST)
#  1622, Tue 13 Apr 10 (PDT)
#  1425, Thu 31 Dec 09 (NZDT)
#
#  iatmon.rb: OWT monitor, with C source classifier
#
#  Copyright 2009-2012, Nevil Brownlee, U Auckland | CAIDA | WAND
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


require "#{File.dirname($0)}/owt-config"

begin
  require 'IPprefix'  # Is it installed so Ruby can find it?
rescue LoadError
  require "#{Home}/IPprefix"  # No, try for it in Home directory
end

require "#{Home}/owt-c/OWTmonitor"  # Must follow IPprefix (!)
require "#{Home}/owt-rb/owt-support" 
require "#{Home}/owt-rb/owt-monitor" 
require "#{Home}/owt-rb/distrib"

require "#{Home}/write-sum"

# ./owtmon.rb dag:/dev/dag0 sum2                                  # ua-live
# ./owtmon.rb traces/ua.20091201.1700.bpf.gz sum2 sum2/san-trace  # ua-trace
# ./owtmon.rb /work/telescope/live/2009-12-21/net44.1261404000.pcap.gz sum2 sum2/test/2009-12-21.fdf  # odin

# ./newmon.rb pcapfile:/home/nevil/ua64-traces/test20101213.011030 sum2
# ./newmon.rb pcapfile:/scratch/owtmon/net44.1299596400.pcap.gz sum2 test.fdf 3 D

trace_URI = ARGV[0]    # URI of trace to process
output_dn = ARGV[1]    # output sum2 directory 
log_fn = ARGV[2]       # name of owtmon log file (appended to)

$clog_level = ARGV.size < 4 ? 3 : ARGV[3].to_i
  # 0 or empty = display all log messages
  # 1          = display Time messages
  # 2          = display Time and Stats messages

distribs_reqd = "-u#{ARGV[4]}" if !ARGV[4].nil?
  # D to write distributions, one for each one-way source
  # T to only write types+groups file (no sum2 file)

trace_URI = SourceURI  if trace_URI.nil?
output_dn = SummaryDir if output_dn.nil?
log_fn = StatsFile     if log_fn.nil?

OWTmonVer = '2.1';  FormatVer = '2.1'

TCP = 6;  UDP = 17;  ICMP = 1;  ICMP6 = 58  # Protocol numbers
ICMP_DEST_UNREACH  =   3  # ICMP Types
ICMP_TTL_EXCEEDED  =  11
ICMP6_DEST_UNREACH =   1  # ICMP6 Types
ICMP6_TTL_EXCEEDED =   3
# TCP flags  32 URG, 16 ACK, 8 PSH, 4 RST, 2 SYN, 1 FIN 

def file_dn(output_dn, ts)
  check_dir_exists(output_dn)
  odn = "#{output_dn}/#{utc_hts(ts)}"
  check_dir_exists(odn)
  odn
end

def day_change(ts)
  if $today_date.nil?
    $today_date = utc_hts(ts)
    return true
  else
    new_date = utc_hts(ts)
    change = new_date != $today_date
    $today_date = new_date
    return change
  end
end

def next_uri(uri)  # Get filename for next telescope trace file
  uri =~ /^(.+):(.+)$/
  utype = $1;  fn = $2
  arg_fn = File.basename(fn)
  
  fn =~ /^(.+)\/(20.+)\/(.+)\.(\d+)\.(.+)$/
  live_dir = $1;  sdate = $2;  net = $3;  ut = $4;  ftype = $5

  return_next = false
  Dir.glob("#{live_dir}/????-??-??").sort.each do |an|
    date = File.basename(an) 
    next if date < sdate
    Dir.glob("#{live_dir}/#{date}/#{net}.*.#{ftype}").sort.each do |fulln| 
      fn = File.basename(fulln)
      return "#{utype}:#{fulln}" if return_next
      return_next = true if fn == arg_fn
    end
  end
end

class NbrHash < Hash  # For port numbers and ip lengths
  def update(str)
    if !self[str]
      self[str] = 1
      return true  # new
    else
      self[str] += 1
      return false
    end
  end

  def val_range
    min = 65535;  max = 1
    self.each do |key, val|
      ikey = key.to_i
      min = ikey if ikey < min
      max = ikey if ikey > max
    end
    [self.size, min, max]
  end

  def mode
    shash = self.sort { |kva, kvb| kvb[1] <=> kva[1] }
    shash[0][0].to_i
  end

  def write(f)
    shash = self.sort { |kva, kvb| kvb[1] <=> kva[1] }
    n = 0
    shash.each do |kv|
      f.printf("  %u %u", kv[0], kv[1])
      n += 1
      break if n == N_ProbePorts
    end
  end

  def write_str(f, n_reqd)
    shash = self.sort { |kva, kvb| kvb[1] <=> kva[1] }
    n = 0
    shash.each do |kv|
      f.print"  #{kv[0]} #{kv[1]}"
      n += 1
      break if n == n_reqd
    end
  end
end  # NbrHash

N_source_types = SrcTypeNames.size
N_source_groups = GroupNames.size

#class OWTevent
#  attr_reader :id
#end

class OWTmessage
  attr_reader :type, :source, :text
end

# sudo dagfour -d /dev/dag0 -aei
# sudo dagfour slen=80 varlen
# sudo dagfour -d /dev/dag0 disableb
# sudo dagfour -d /dev/dag0 enablea  # A interface sees UA border traffic
# sudo dagsnap -v -o tracefile  # Check it's working

Type_s = {OQ_TXTMSG => 'OQ_TXTMSG', 
  OQ_TIMED_OUT => 'OQ_TIMED_OUT',
  OQ_END_OF_FILE => 'OQ_END_OF_FILE' }

$stderr.print \
  "trace_URI=#{trace_URI}, output_dn=#{output_dn}, log_fn=#{log_fn}\n"

$owt = Monitor.new("-i#{trace_URI} -s#{N_Sources} -n#{MinPkts} -t#{N_source_types} -c#{Hours_reqd*3600} -h#{Hours_reqd} -l#{MeterLoc} -o#{log_fn} #{distribs_reqd}")
addr_list = Array.new
v6_in_list = false

LocalAddress.split(/,\s*/).each do |la|
  as = IPprefix.from_s(la)
  print "as = #{as}, ver=#{as.version}, len=>#{as.length}\n";  $stdout.flush

  addr_list.push(IPprefix.from_s(la))
  v6_in_list = true if la =~ /:/
end

LocalV6 = v6_in_list
$owt.set_local_addrs( addr_list )

(Meter_ver, TwoWay) = $owt.info
print "Meter_ver=#{Meter_ver}, TwoWay=#{TwoWay}\n"

LiveTrace = $owt.live
if LiveTrace
  Trace_sz = Host  # Dummy for trace from live interface
else
  if $owt.n_interfaces != 1
    $stderr.print "Can only process one file per owtmon.rb run!\n"
    exit
  else
    Trace_sz = sprintf("%.3f", File.size($owt.last_fn)/1000000000.0)  # GB
  end
end

def sigint
  print "^C pressed: stopping iatmon ...\n";  $stdout.flush
  $owt.close
end

#trap(:INT, lambda {sigint})  # ^C (Interrupt) =>  Set request_shutdown [1.8]
trap(:INT) { sigint }  # ^C (Interrupt) =>  Set request_shutdown [1.9]

#def sigtstp
#  $stderr.print "SIGTSTP (^Z pressed) ..\n"
#  st = caller(3)  # Ignore first 3 lines
#    # First line printed is where the signal found us
#  $stderr.print st.join("\n")
#  $stderr.print "\n"
#end 
##trap(:TSTP) { sigtstp }  # ^Z (Temporary stop) => show traceback

trap(:TSTP) { }  # Ignore ^Z (used to stop mk-hourly-sum2.rb)
  # NOTE: ^Z won't push owtmon.rb to background - ask it to stop with ^C
  #       or force it to stop with KILL -9 (after ps to get its procid)

$start_proc_time = Process::times;  $start_proc_date = Time.now

np_dist = nil

MkStrmSzDistribs = false  # Don't write stream-size distribs
szf = File.new('size.dat', 'w') if MkStrmSzDistribs

if Hours_reqd > 1  # Assume we want to process n hourly trace files
  uri = "pcapfile:#{trace_fn}"
  nu = 0;  other_uris = Array.new
  (1..Hours_reqd-1).each do
    new_uri = next_uri(uri)
    other_uris.push(new_uri)
    uri = new_uri
  end
  $owt.set_uris(other_uris)
end

n = 0
first_time = true
sum_start_ts = $owt.open
if LiveTrace
  log_fn = "#{file_dn(output_dn, sum_start_ts)}/#{utc_gts(sum_start_ts)}.fdf"
end

$owt.each do |msg|
  # print "#{Type_s[msg.type]} =====\n"
  case msg.type
  when OQ_TXTMSG
    type = msg.text[1..4]
    exit if msg.text =~ /Fatal/
    mts = msg.source  # g_last_ts for TXTMSGs
    case type  # txt message type
    when 'Time'  #Time
      $owt.log_msg(-1, msg.text)  # Write to log file
      $stdout.flush
      if MkStrmSzDistribs
        szf.print "#{utc_gts(mts)} "
        if first_time
          np_dist = Distribution.new(1, 400, 100, DT_LOG, 1.0)
          np_dist.write_x(szf, false)
          first_time = false
        else
          $owt.each_source_size do |sz|
            np_dist.count(sz)
          end
          np_dist.write_y(szf, false)
          np_dist = Distribution.new(1, 400, 100, DT_LOG, 1.0)
        end
      end
    when 'Reco'  # Record: time to write sum2 file
      fn = "#{file_dn(output_dn, sum_start_ts)}" +
        "/#{utc_gts(sum_start_ts)}.#{SumFileType}"
      if distribs_reqd.nil?  # Not writing distribution info file
        write_sum_file(fn, sum_start_ts, mts) 
      end
      if day_change(mts)  # Start new log file
        log_fn = "#{file_dn(output_dn, mts)}/#{utc_gts(mts)}.fdf"
        $owt.new_log_file(log_fn)
      end
      sum_start_ts = mts
    when 'Stat', 'SrcS', '#iat'  # Stats record, startup message
      $owt.log_msg(-1, msg.text)  # Write to log file
      $stdout.flush
    when 'Shut'  #Shutdown
      $owt.log_msg(-1, msg.text)  # Write to log file
      $stdout.flush  # and console
      if MkStrmSzDistribs
        szf.print "#{utc_gts(mts)} "
        $owt.each_source_size do |sz|
          np_dist.count(sz)
        end
        np_dist.write_y(szf, false)
      end
      fn = "#{file_dn(output_dn, sum_start_ts)}" +
        "/#{utc_gts(sum_start_ts)}.#{SumFileType}"
      write_sum_file(fn, sum_start_ts, mts)
        # $owt.each terminates after the #Shutdown message !?
    else  # Shouldn't happen
      $owt.log_msg(-1, "#{msg.text} <<<<<")  # Write to log file
      $stdout.flush  # Always written to console
    end
  when OQ_END_OF_FILE, OQ_TIMED_OUT  # Don't expect to see these in v2.0
    print "#{Type_s[msg.type]} ---\n"
    event = msg.source
  else  # Shouldn't happen !
    event = msg.source
    $owt.log_msg(0, "#{Type_s[msg.type]}:  ")
    first_fwd, src_addr, tot_pkts, lifetime = event.usage
    inward = first_fwd ? '->' : '<-'
    $owt.log_msg(-1, sprint("   %s %s  %d  %.6f *****", src_addr, inward, tot_pkts, lifetime))
    $stdout.flush
    exit
  end
#  n += 1
#    exit if n == 5
end

#szf.close if MkStrmSzDistribs

$fdf.close
