#!/usr/bin/env ruby

# 1648, Sat 11 Feb 12 (CET)
# 1635, Mon  4 Jan 10 (NZDT)
#
# mk-iou1-hourly-sum2.rb: runs through UNINETT trace files
#   producing hourly statistics using iatmon.rb
#
# For now, just write summary files into SummaryDir directory.
#   Will make an sql database from them one day.
#
# Copyright 2009-2012, Nevil Brownlee, UNINETT | U Auckland

ruby_bin = '/home/nevil/.rvm/rubies/ruby-1.9.2-p136/bin/ruby'
  # ml-hourly will be executed in root environment,
  #   which doesn't have rvm

require "#{File.dirname($0)}/owt-config"
Live = '/data/hint'
SumFiletype = 'sum2'

require "#{Home}/owt-rb/owt-support" 

#require 'tzinfo'
#ServerCountry = 'America/Los_Angeles'  # Telescope timezone
#Ttz = TZInfo::Timezone.get(ServerCountry)

# Telescope times are UTC, but Ruby gives them to us in local time

def utc_gts(local_ts)  # Convert ts to gnuplot timefmt
  ds = local_ts.to_i+30  # strfmtime will round down to closest hour
#  begin
#    Time.at(Ttz.local_to_utc(ds).to_i).strftime("%Y%m%d.%H%M")
    Time.at(ds).gmtime.strftime("%Y%m%d.%H%M")
      # Assume that gmtime uses TZ environment variable
      # time format suitable for gnuplot !
#  rescue
#    "20091101.0100"  # Dummy to handle 'ambiguous time' at PDT changeover
#  end
end

def set_start_date
  tn = Time.now.to_i 
  Time.at(tn - 4*24*3600).strftime('%Y-%m-%d') 
end

if ARGV.size == 0
  start_date = set_start_date()
elsif ARGV[0] == 'F'
  start_date = set_start_date()
else
  start_date = ARGV[0]
  last_date = ARGV[1]  # nil if only one arg
end
print "start_date=#{start_date}, last_date=#{last_date}.\n"

$break_loop = false
def sigtstp
  $stderr.print "  mk-hourly-sum2.rb will stop after current file\n"
  $break_loop = true
end 
trap(:TSTP) { sigtstp }  # ^Z (Temporary stop)

trap(:INT) { }  # Ignore ^C (it's children will see the :INT)

sum_dir = "#{Home}/#{SummaryDir}"
check_dir_exists(sum_dir)

nf = 0
Dir.glob("#{Live}/*.pcap").sort.each do |fn|
  fnb = File.basename(fn, '.pcap')

  t = Time.gm(fnb[0..3].to_i, fnb[4..5].to_i, fnb[6..7].to_i,
    fnb[8..9].to_i)
  start_t = Time.at((t.to_i - 3580)/3600*3600).gmtime
  print "#{t}, starts at #{start_t} <<<\n"
  start_h = "#{start_t.strftime('%H')}00"
  start_d = start_t.strftime('%Y-%m-%d')
  short_start_d = start_t.strftime('%Y%m%d')

  # print "  date=#{start_d}, date<start_date=#{start_d < start_date}\n"
  next if start_d < start_date
  break if !last_date.nil? && start_d > last_date

  trace_sz = sprintf("%.3f", File.size(fn)/1000000000.0)  # GB
  print "  fn=#{fn}, starts at #{start_d}.#{start_h}, trace_sz=#{trace_sz}\n"

  check_dir_exists("#{sum_dir}/#{start_d}")
  op_prefix = "#{sum_dir}/#{start_d}/#{short_start_d}.#{start_h}"
  log_fn = "#{op_prefix}.fdf"
  sum_fn = "#{op_prefix}.#{SumFiletype}"

  have_sum_file = already_there(sum_fn)
  print "sum_fn=#{sum_fn}, have_sum_file=#{have_sum_file}\n"

  if !already_there(sum_fn)
    cmd = "#{ruby_bin} #{Home}/iatmon.rb pcapfile:#{fn} #{sum_dir} #{log_fn} 0"
      # 0 => display all log messages
    print "cmd: #{cmd}\n"
    system(cmd)
    # exit  # Just do one file !!!
    exit if $break_loop
  end

  nf += 1  #;  break if nf == 5
end
