#!/usr/bin/env ruby

# 1039, Tue 24 May 11 (NZST)
#
# mon-check-sum2.rb: Report on a Monitor's sum2 files
#   ./mon-check-sum2.rb  2011-01-03 2011-05-31
#
# Copyright 2009-2012, Nevil Brownlee, CAIDA | U Auckland

require "#{File.dirname($0)}/owt-config"
require "#{Home}/owt-rb/owt-support"  # owtmon utilities and constants

StatsDir = SummaryDir

def print_error(msg)
  $stderr.print ">>> #{msg}\n"
  $stderr.print "    #{$0}  monitor  first-ymd last_ymd  reqd_hour\n"
  exit
end

start_date = ARGV[0]
if start_date =~ /^(20\d\d)-(\d\d)-(\d\d)$/
   sd = $1 + $2 + $3
else 
  print_error("first_date not yyyy-mm-dd !!")
end
last_date = ARGV[1]
if last_date =~ /^(20\d\d)-(\d\d)-(\d\d)$/
   ed = $1 + $2 + $3
else 
  print_error("last_date not yyyy-mm-dd !!")
end
date_range = "#{sd}-#{ed}"

sds = SummaryDataSet.new(StatsDir, start_date,last_date, false)

last_ts = last_format_ver = nil
sds.each do |sf|  # Get files within range of days
  #print "#{sf.ymdh} ...\n"
  err = late_start = false

  if sf.format_ver != last_format_ver
    if !last_format_ver.nil?
      print "#{sf.ymdh}\n" if !err;  err = true
      print "  Format change, to #{last_format_ver}\n"
    end
    last_format_ver = sf.format_ver
  end

  sf.ymdh =~ /(\d\d)(\d\d)(\d\d)\.(\d\d)(\d\d)/
  minutes = $5.to_i
  ts = Time.gm($1, $2, $3, $4, $5, 0, 0).to_i

  if minutes != 0
    print "#{sf.ymdh}\n" if !err;  err = true
    print "  Not the start of an hour <<<\n"
    late_start = true
  end

  if !last_ts.nil?
    seconds_since_last = ts - last_ts
    if seconds_since_last < 3600 && !late_start
      print "#{sf.ymdh}\n" if !err;  err = true
      print "  Late start <<<\n"
    elsif seconds_since_last > 3600
      print "#{sf.ymdh}\n" if !err;  err = true
      print "  Gap (#{seconds_since_last}) > 3600 s\n"
    end
  end
  last_ts = ts

  tl = sf.find_section('time')  # #times (minutes)
  if tl =~ /duration=(.+),/
    duration = $1.to_f
    if duration < 59.95 && !late_start
      print "#{sf.ymdh}\n" if !err;  err = true
      print "  Duration #{duration}, too short\n"
    end
  else
    print "#{sf.ymdh}\n" if !err;  err = true
    print "  Couldn't get duration ***\n"
  end

end
# mb_f.close

