/* 1541, Thu 17 Mar 11 (PDT)
   1552, Thu 17 Jun 10 (NZST)
   1955, Thu 22 Apr 10 (PDT)
   1330, Fri  1 Feb 08 (NZDT)

   owt_watcher.c:  Packet and Source routines for owtmon

   iatmon/owtmon: One-Way Traffic Monitor testbed
   Copyright (C) 2009-2012 by Nevil Brownlee, U Auckland | CAIDA | WAND

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "owt.h"

/* Globals for SRI code to generate Conficker port numbers (in conf-sri.c) */
typedef union {
   int16_t s16[8];  uint16_t u16[8];  int32_t s32[4];
   } result_t;
int portgen(int ip, result_t *res, int week);

uint32_t date_seed;

void set_date_seed(uint32_t ts) {
   /* Emile's python code ...
      date_seed = ((((uint64_t)ts - 345600) * 1861606989) >> 32) >> 18; */
   /* Nevil's C code: following line seems to be equivalent to
      date_seed = ((uint64_t)ts - 4*24*3600)/(7*24*3600); */
   date_seed = (uint32_t)(((((uint64_t)ts - 345600) * 1861606989) >> 32) >> 18);
   //d printf("   l_t_s=%u, date_seed=%x\n", ts, date_seed);
   }

static char const *bT_DHT_val = "d1:ad2:id20:";  /* 12 bytes 0..11 */
static char const *uT_proto_val =  /* 16 bytes 12..27 */
   "\x7f\xff\xff\xff\xab\x02\x04\x00\x01\x00\x00\x00\x08\x00\x00\x00";

#define test_ut \
   pp->uTorrent = \
      (hl >= dx+8+0+12 && \
         (memcmp(&hdr[dx+8+0], bT_DHT_val, 12) == 0)) || \
      (hl >= dx+8+12+16 && \
       (memcmp(&hdr[dx+8+12], uT_proto_val, 16) == 0));

uint32_t tcp_key(int dx, unsigned char *hdr, int hl) { 
   uint32_t val;  int x;

   if (hl < dx+16) return 0;
   if ((hdr[dx+13] & 0x02) == 0)  /* SYN flag not present */
      return 0;
   val = ntohs(*(uint16_t *)&hdr[dx+14]);  /* Window size */

   if ((hdr[dx+12] >> 4) == 5) return val;  /* No options */

   x = dx+20;
   for (;;) { 
      switch (hdr[x]) { 
      case 0:  /* End of option list */ 
         return val; 
      case 1:  /* No-op */ 
         x += 1; 
         break;
      case 2:  /* MSS */ 
         val |= (ntohs(*(uint16_t *)&hdr[x+2])-600) << 16;  /* 10 bits */
         break;
      case 3:  /* Window scale */ 
         val |= ntohs(*(uint16_t *)&hdr[x+2]) << 28;  /* Top 4 bits */
         break;
      case 4:  /* SACK permitted */
         val |= 0x08000;
         break;
      case 8:  /* Timestamp */
         val |= 0x04000000;
         break;
         }
      if (x+1 >= hl || hdr[x+1] == 0) break; 
      x += hdr[x+1]; 
      if (x >= hl) break; 
      } 
   return val; 
   } 

int get_v4key(struct pkt_info **r_pp, int *r_fwd, int hl, unsigned char *hdr)
{  /* Code from pkt_extract() in NeTraMet's flowhash.c */
   struct pkt_info *pp;
   int ver, dx, frag;
   struct ip_address *saddr, *daddr;  uint16_t sport, dport;
   int sa_local, da_local, fwd;
   uint32_t v4_temp;

   if ((ver = hdr[0] >> 4) == 4) {
      if (hl < 12+IP4_ADDR_LEN*2) {
         *r_pp = NULL;
         return SKEY_NOPEERADDR;  /* Didn't get the peer addresses! */
         }
      if (frag = (ntohs(*(uint16_t *)&hdr[6]) & 0x1FFF) != 0) {
 	 *r_pp = NULL;
         return SKEY_FRAGPKT;  /* Non-first fragment */
         }
      if ((pp = get_pkt_info()) == NULL) return SKEY_NOSPACE;
      pp->version = 4;
      pp->proto = hdr[9]; /* Protocol Type */
      dx = (hdr[0] & 0x0F) << 2;  /* IPv4 header length in 32-bit units */
      pp->tcp_flags = 0;

      saddr = &pp->src_addr;  daddr = &pp->dst_addr;
      saddr->a.v4 = *(uint32_t *)&hdr[12];  /* IPv4 source */
      daddr->a.v4 = *(uint32_t *)&hdr[16];  /* IPv4 dest */
      sa_local = addr_is_local(4, saddr);
      da_local = addr_is_local(4, daddr);
      if (sa_local && !da_local) fwd = 0;
      else if (!sa_local && da_local) fwd = 1;
      else {
         free_pkt_info(pp);  *r_pp = NULL;
         return SKEY_NOLOCALADDR;
         }
      if (!fwd) {
	 v4_temp = saddr->a.v4;  saddr->a.v4 = daddr->a.v4;
         daddr->a.v4 = v4_temp;
         }

      switch (pp->proto) {
      case PT_TCP:
#if HALF_WAY_SOURCES
 	 pp->tcp_seq = ntohl(*(uint32_t *)&hdr[dx+4]);
 	 pp->tcp_ack = ntohl(*(uint32_t *)&hdr[dx+8]);
	    /* We only want to see whether seq or ack is increasing,
               so we don't need to check the ACK flag */
#endif
         pp->tcp_flags = hl >= dx+14 ? hdr[dx+13] : 128;

      case PT_UDP:
         if (hl < dx+4) {
  	    free_pkt_info(pp);  *r_pp = NULL;
            return SKEY_NOTRANSADDR;  /* Didn't get the IP ports! */
            }
         else {
            sport = ntohs(*(uint16_t *)&hdr[dx]);  /* Source port */
            dport = ntohs(*(uint16_t *)&hdr[dx+2]);  /* Dest port */
	    }

         if (fwd) {
            pp->src_port = sport;  pp->dst_port = dport;
            }
         else {
            pp->src_port = dport;  pp->dst_port = sport;
            }
         //      printf("*** src=%08x, dst=%08x, fwd=%d\n",  //p
         //	     sd4p->src_addr, sd4p->dst_addr, fwd);
         break;
      case PT_ICMP:
         pp->src_port = 0;    /* Type field */
         pp->dst_port = ntohs(*(uint16_t *)&hdr[dx]);  /* Type|Code */
         break;
      default:
         pp->src_port = pp->dst_port = 0;
         break;
         }
      //d  printf("get_v4key(1): pktp=%0x, saddr=%x, daddr=%x, pp=%x\n",
      //d     pktp, saddr, daddr, pp);


      pp->ip_len = ntohs(*(uint16_t *)&hdr[2]);
      pp->ttl = hdr[8];
      pp->uTorrent = 0;  pp->tcp_key = 0;
      if (pp->proto == PT_TCP) {
	 pp->tcp_key = tcp_key(dx, hdr, hl);
         }
      else if (pp->proto == PT_UDP) {
         test_ut;
         }
      *r_fwd = fwd;
      *r_pp = pp;  /* Return source (with ptrs to pkt_data and sd_tuple_v4) */
      return SKEY_OK;
      }
   else return SKEY_BADIPHDR;
   }

int find_v6_type(int *prot, int *hx, const uint8_t *p, const int hl)
{
   int nh, nx;

   nh = p[6];  nx = 40;
   for (;;) {
      switch (nh) {
      case PT_6FRAG:    /* Fragment */
      case PT_6ESP:     /* Encapsulating Security Payload (ESP) RFC 2406 */
      case PT_6NONH:    /* No next header */
         *prot = nh;
         return 0;  /* Can't analyse this packet */

      case PT_6HOP:     /* Hop-by-hop options */
      case PT_6ROUT:    /* Routing (Type 0) */
      case PT_6DEST:    /* Destination options */
      case PT_6AH:      /* Authentication (AH) RFC 2402 */
      case PT_6MOBILE:  /* IPv6 Mobility Header */
	 break;  /* Keep looking */

      default:
         *prot = nh;  *hx = nx;
         return 1;  /* Can see into this one */
         }
      if (nx >= hl) {
	 *prot = -1;  /* Not enough header bytes to get Next Header */
         return 0;
         }
      nh = p[nx];
      nx += (p[nx+1]+1) * 8;
      }
   }

int get_v6key(struct pkt_info **r_pp, int *r_fwd, int hl, unsigned char *hdr)
{  /* Code from pkt_extract() in NeTraMet's flowhash.c */
   struct pkt_info *pp;
   struct ip_address *saddr, *daddr;  uint16_t sport, dport;
   int ver, dx, prot;
   int sa_local, da_local, fwd;
   struct v6_address temp;

   if ((ver = hdr[0] >> 4) == 6) {
      if (hl < 8+IP6_ADDR_LEN*2) {
         *r_pp = NULL;
         return SKEY_NOPEERADDR;  /* Didn't get the peer addresses! */
         }
      if (!find_v6_type(&prot,&dx, hdr,hl)) {
         *r_pp = NULL;
         if (prot < 0)  /* Not enough bytes to determine protocol */
  	    return SKEY_V6SMALLSNAP;
         else return SKEY_V6OPAQUE;  /* Can't see into packet */
         }

      if ((pp = get_pkt_info()) == NULL) return SKEY_NOSPACE;
      pp->version = 6;
      saddr = &pp->src_addr;  daddr = &pp->dst_addr;
      memcpy(saddr->a.a, &hdr[8], IP6_ADDR_LEN);  /* IPv6 source */
      memcpy(daddr->a.a, &hdr[24],IP6_ADDR_LEN);  /* IPv6 dest */
      sa_local = addr_is_local(6, saddr);
      da_local = addr_is_local(6, daddr);
      if (sa_local && !da_local) fwd = 0;
      else if (!sa_local && da_local) fwd = 1;
      else {
    	 free_pkt_info(pp);  *r_pp = NULL;
         return SKEY_NOLOCALADDR;
         }
      if (!fwd) {
         memcpy(&temp, saddr, sizeof(struct v6_address));
         memcpy(saddr, daddr, sizeof(struct v6_address));
         memcpy(daddr, &temp, sizeof(struct v6_address));
         }
      switch (prot) {
      case PT_TCP:
#if HALF_WAY_SOURCES
 	 pp->tcp_seq = ntohl(*(uint32_t *)&hdr[dx+4]);
 	 pp->tcp_ack = ntohl(*(uint32_t *)&hdr[dx+8]);
	    /* We only want to see whether seq or ack is increasing,
               so we don't need to check the ACK flag */
#endif
         pp->tcp_flags = hl >= dx+14 ? hdr[dx+13] : 128;
      case PT_UDP:
         if (hl < dx+4) {
            free_pkt_info(pp);  *r_pp = NULL;
            return SKEY_NOTRANSADDR;  /* Didn't get the IP ports! */
            }
         else {
            sport = ntohs(*(uint16_t *)&hdr[dx]);  /* Source port */
            dport = ntohs(*(uint16_t *)&hdr[dx+2]);  /* Dest port */
            }
         if (fwd) {
            pp->src_port = sport;  pp->dst_port = dport;
            }
         else {
            pp->src_port = dport;  pp->dst_port = sport;
            }
         break;
      case PT_6ICMP:
         pp->src_port = 0;    /* Type field */
         pp->dst_port = ntohs(*(uint16_t *)&hdr[dx]);  /* Type|Code */
         break;
      default:
         pp->src_port = pp->dst_port = 0;
         break;
         }

      pp->ip_len = 40 + ntohs(*(uint16_t *)&hdr[4]);
         /* Header + payload (includes extension headers) */
      pp->ttl = hdr[7];

      if (pp->proto == PT_TCP) {
	 pp->tcp_key = tcp_key(dx, hdr, hl);
         }
      else if (pp->proto == PT_UDP) { 
         test_ut;
         }

      *r_fwd = fwd;
      *r_pp = pp;  /* Return source (with ptrs to pkt_data and sd_tuple_v6) */
      return SKEY_OK;
      }
   else return SKEY_BADIPHDR;  /* Wasn't a v6 packet! */
   }

#if 0
void print_source(struct source *sp) {
   if (sp->version != 4) printf("source: version=%d <<<\n", sp->version);
   else {
      uint8_t *a = sp->src_addr.a.a;
      printf("%u.%u.%u.%u", a[0],a[1],a[2],a[3]);
      a = sp->d.si->dst_addr.a.a;
      printf("->%u.%u.%u.%u", a[0],a[1],a[2],a[3]);
      }
   }
#endif

void check_src_nulls(struct source_queue *sq, struct source *sp);
int check_source_chains(char const *msg, struct source_queue *sq);

void enqueue_source_status(int type, struct source *sp, uint32_t now_ts) {
   struct opq_entry *oqe;
   t_semaphore_down(&opq_empty);  /* Wait for an empty entry */
   oqe = oqe_for_source(sp);
   oqe->type = type;
   opq_entry_q_lock(&opq);
   opq_entry_push(&opq, oqe);
   opq_entry_q_unlock(&opq);
   sp->d.si->last_report_sec = now_ts;
   t_semaphore_up(&opq_ready);  /* Signal a full entry */
   }

void enqueue_source_txt(char *msg) {
   struct opq_entry *oqe;
   t_semaphore_down(&opq_empty);  /* Wait for an empty entry */
   oqe = oqe_for_source(NULL);
   oqe->type = OQ_TXTMSG;  oqe->msg = msg;
   oqe->ts = g_last_ts;
   opq_entry_q_lock(&opq);
   opq_entry_push(&opq, oqe);
   opq_entry_q_unlock(&opq);
   t_semaphore_up(&opq_ready);  /* Signal a full entry */
   }

static int rake_idle_sources(struct flow_watcher_control *fw, uint32_t now_ts) {
   struct source_queue *sq = &fw->sq;
   double min_idle_s, last_t_s;
   struct source *sp;
   int n, keep, examined = 0, freed = 0;

#if SC_DEBUG >= 2
   fprintf(stderr, "  === check_timeouts(%d), ts=%u\n", fw->id, now_ts);
#endif
   last_t_s = now_ts - start_ts;
   min_idle_s = last_t_s - TOO_MANY_TIMEOUT;
#if SC_DEBUG >= 1
   check_source_chains("--- starting rake_idle_streans", sq);
#endif
   source_q_lock(sq);
   for (n = sq->length; n != 0;  --n) {  /* Don't go round more than once */
      sp = source_shift(sq);  /* Remove source from head */
      if (sp->d.si->last_s > min_idle_s) {
 	 source_unshift(sq, sp);  /* Put it back! */
         break;  /* Hasn't been queued long enough; stop checking */
         }
      else {  /* Could be inactive, can we recover it? */
	 examined += 1;
	 keep = sp->totp > MN_PKTS_FOR_DISCARD;
         }
      if (keep) source_push(sq, sp);  /* Add source to tail */
         /* We won't waste time tesing it until it drifts back
            to the head of the queue.  That means we no longer
            have the queue in (approximnate) LastTime order */
      else {  /* Discard this source */
	 update_ua_stats(sp);  /* Un-analysed source info */
         update_summary_info(sp); 
         owt_ht_delete(fw->sh_table, sp);  /* Remove from hash table */
	 sp->next = sp->prev = (struct source *)0x55555555;
         free_source(sp);
	 freed += 1;
	 if (freed == MN_SOURCES_TO_FREE) break;
	 //	 check_source_chains("--- after timed out", sq);
         }
      }
#if SC_DEBUG >= 2
   n = source_space.mx_objects - source_space.free_queue.length;
   fprintf(stderr, "   ==== sources: %d examined, %d freed.  %d active sv4s\n",
      examined, freed, n);
#endif

   source_q_unlock(sq);
   return freed;
   }

int coss(struct source *sp)
{  /* Return 0 to continue, 1 to stop walking */
   struct flow_watcher_control *fw = 
      &flow_watcher_control_space.objects[sp->fw_nbr];
   fw->state_counts[sp->state] += 1;
   return 0;
   }

void count_source_states(char *buf, struct flow_watcher_control *fw,
      const char *msg) {
  int j, x, y;
   for (j = 0; j != SS_n_states; j += 1) fw->state_counts[j] = 0;
   source_walk_up(&fw->sq, coss);
   source_walk_up(&fw->sq2, coss);
   x = sprintf(buf,"%s %d ", msg, fw->id);
   for (j = SS_ONE_WAY; j != SS_n_states; j += 1) {
     y = sprintf(&buf[x], " %d", fw->state_counts[j]);
     x += y;
     }
   }

void check_was_two_way_timeouts(
   struct flow_watcher_control *fw, uint32_t now_ts) {
   struct source_queue *sq2 = &fw->sq2;
   uint32_t last_t_s,  min_idle_s;
   struct source *sp;
   int keep, terminated, n, ow_terminated;
   int counted = 0, freed = 0;

   last_t_s = now_ts - start_ts;
   min_idle_s = last_t_s - WAS_TWO_WAY_TIMEOUT;
#if  SC_DEBUG >= 2
   printf("  =2= check_wtw_timeouts(%d), ts=%u, last_ts=%u, min_idle=%u\n",
      fw->id, now_ts, last_t_s, min_idle_s);
#endif
   ow_terminated = terminated = 0;
#if SC_DEBUG >= 1
   check_source_chains("--- starting check_was_two_way_timeouts", sq2);
#endif
   source_q_lock(sq2);
   for (n = sq2->length; n != 0;  --n) {  /* Don't go round more than once */
      counted += 1;
      sp = source_shift(sq2);  /* Remove source from head */
      if (sp->state == SS_DELETE) keep = 0;
      else if (sp->ts > min_idle_s) {
 	 source_unshift(sq2, sp);  /* Put it back! */
         break;  /* Hasn't been queued long enough; stop checking */
         }
      else {  /* Could be inactive, can we recover it? */
	 keep = sp->state != SS_WAS_TWO_WAY;  /* Inactive too long */
         }
      if (keep) source_push(sq2, sp);  /* Add source to tail */
      else if (sp->state == SS_DELETE) {
         terminated += 1;
         sp->state = SS_TIMEOUT;
         owt_ht_delete(fw->sh_table, sp);  /* Remove from hash table */
	 sp->next = sp->prev = (struct source *)0x55555555;
         free_source(sp);
	 freed += 1;
	 }
      else {  /* Timed-out source */
         terminated += 1;
         update_summary_info(sp); 
         sp->state = SS_TIMEOUT;
         owt_ht_delete(fw->sh_table, sp);  /* Remove from hash table */
	 sp->next = sp->prev = (struct source *)0x55555555;
         free_source(sp);
	 freed += 1;
	 //	 check_source_chains("--- after timed out", sq2);
         }
      }
#if SC_DEBUG >= 2  /* (will print every second!) */
   printf("   =2= %d sources terminated, %d freed, %d counted.  %d WAS_TWO_WAY sources\n",
      terminated, freed, counted, sq2->length);
   fflush(stdout);
#endif
   source_q_unlock(sq2);
   }

void check_source_timeouts(struct flow_watcher_control *fw, uint32_t now_ts) {
  struct source_queue *sq = &fw->sq, *sq2 = &fw->sq2;
   double min_idle_s2, min_idle_s1, last_t_s, av_int;
   struct source *sp;
   int keep, terminated, n, ow_terminated;
   int counted = 0, freed = 0;

#if SC_DEBUG >= 2
   printf("  === check_timeouts(%d), ts=%u\n", fw->id, now_ts);
#endif
   // char cos_buf[120];
   // count_source_states(cos_buf, fw, "Starting check_source_timeouts");
   // printf(cos_buf);  fflush(stdout);
   last_t_s = now_ts - start_ts;
   min_idle_s1 = last_t_s - ONE_WAY_TIMEOUT;
   min_idle_s2 = last_t_s - TWO_WAY_TIMEOUT;
   ow_terminated = terminated = 0;
#if SC_DEBUG >= 1
   check_source_chains("--- starting check_source_timeouts", sq);
#endif
   source_q_lock(sq);  source_q_lock(sq2);
   for (n = sq->length; n != 0;  --n) {  /* Don't go round more than once */
      counted += 1;
      sp = source_shift(sq);  /* Remove source from head */
      if (sp->state == SS_ONE_WAY ||
#if HALF_WAY_SOURCES 
            sp->state == SS_HALF_WAY ||
#endif
            sp->state == SS_NEW_ONE_WAY)
         keep = 1;
      else if (sp->state == SS_DELETE) keep = 0;
      else if (sp->d.si->last_s > (sp->d.si->timeout == STO_ONE_WAY ?
            min_idle_s1 : min_idle_s2) ) {
 	 source_unshift(sq, sp);  /* Put it back! */
         break;  /* Hasn't been queued long enough; stop checking */
         }
      else {  /* Could be inactive, can we recover it? */
	if (sp->totp == 1) keep = 1;  /* Only one pkt seen */
        else {
	    av_int = (sp->d.si->last_s - sp->d.si->first_s)/
               (double)(sp->totp - 1);
            keep = ((last_t_s - sp->d.si->last_s)  /* Observed idle us */
		    < av_int*STO_MULTIPLIER);  /* Current timeout estimate */
	    }
         }
      if (keep)  /* Don't waste time tesing it until it drifts back
            to the head of the queue.  That means we no longer
            have the queue in (approximnate) LastTime order */
         source_push(sq, sp);  /* Add source to tail */
      else if (sp->state == SS_DELETE) {
         terminated += 1;
         sp->state = SS_TIMEOUT;
         owt_ht_delete(fw->sh_table, sp);  /* Remove from hash table */
	 sp->next = sp->prev = (struct source *)0x55555555;
         free_source(sp);
	 freed += 1;
	 }
      else {  /* Timed-out source */
	 if (sp->state == SS_TWO_WAY) {  /* Will it reappear as two-way? */
            sp->state = SS_WAS_TWO_WAY;
            if (sp->totp > MN_PKTS_FOR_TWO_WAY)  /* Probably two_way */
	       free_source_info(sp);
	    }
         else log_msg(LOG_ERR, 0, "Time-out, source state = %d", sp->state);
         source_push(sq2, sp);  /* Move it to timed_out queue */
	 /* check_source_chains("--- after timed out", sq); */
         }
      }
#if SC_DEBUG >= 2  /* (will print every second!) */
   printf("   ==== %d sources terminated, %d freed, %d counted.  %d active sources\n",
      terminated, freed, counted, sq->length);
#endif
   source_q_unlock(sq);  source_q_unlock(sq2);
   }

uint32_t swp_now_ts;
int get_source_with_pkts(struct source *sp)
{  /* Return 0 to continue, 1 to stop walking */
   if (sp->state == SS_ONE_WAY)  /* One-way source */
      enqueue_source_status(OQ_END_OF_FILE, sp, swp_now_ts);
   return 0;
   }

void output_active_sources(struct source_queue *sq, uint32_t now_ts) {
   source_q_lock(sq);
   check_source_chains("--- starting output_active_src_hosts",sq);
   swp_now_ts = now_ts;
   source_walk_up(sq, get_source_with_pkts);
   check_source_chains("+++ leaving output_active_src_hosts", sq);
   source_q_unlock(sq);
   }

void check_src_nulls(struct source_queue *sq, struct source *sp) {
   if (sp->next == NULL && sp != sq->tail) {
      printf("null sp.next and sp != sp tail ***\n");
      quack(991);
      }
   if (sp->prev == NULL && sp != sq->head) {
      printf("null sp.prev and sp != s head ***\n");
      quack(992);
      }
   }

static char ftb[32];
char *fmt_time(int utc_time, uint32_t *ti) {  /* From NeTraMet */
#if SIZEOF_TIME_T == 8
   uint64_t lti = *ti;
#else
   uint32_t lti = *ti;
#endif
   if (!utc_time) {
      char *ts = ctime((time_t *)&lti);
      strncpy(&ftb[0],  &ts[11], 9);  /* 17:31:42_ */
      strncpy(&ftb[9],  &ts[0],  4);  /* Thu_ */
      strncpy(&ftb[13], &ts[8],  3);  /* 23_  */
      strncpy(&ftb[16], &ts[4],  4);  /* Sep_ */
      strncpy(&ftb[20], &ts[20], 4);  /* 1997 */
      ftb[24] = '\0';
      }
   else {
      struct tm *t = gmtime((time_t *)&lti);
      char const *dow[] = {
         "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
      char const *moy[] = {
         "Jan", "Feb", "Mar", "Apr", "May", "Jun",
         "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
      sprintf(ftb, "%02d:%02d:%02d %s %2d %s %4d",
         t->tm_hour,t->tm_min,t->tm_sec, dow[t->tm_wday],
	 t->tm_mday, moy[t->tm_mon], t->tm_year+1900);
      }
   return ftb;
   }   

static uint32_t last_timerec_sec;
static void output_time_record(uint32_t now_ts) {
   struct source_txt *st = get_source_txt_object(&source_txt_space);
   if (st != NULL) {
      sprintf(st->msg, "#Time: %s  %s  Flows from %lu to %lu",
         fmt_time(1, &now_ts), trace_name,
         (unsigned long)last_timerec_sec, (unsigned long)now_ts);
      enqueue_source_txt(st->msg);
      }
   last_timerec_sec = now_ts;
   }

static void fdf_hdr_string(char *buf) {
   sprintf(buf, "##iatmon-%s: -c%u %s -s%u  starting at %s",
      OWTversion, time_rec_interval, trace_name,
      source_space.mx_objects, fmt_time(1, &start_ts));
   }

static void output_fdf_hdr(void) {
   struct source_txt *st = get_source_txt_object(&source_txt_space);
   if (st != NULL) {
      fdf_hdr_string(st->msg);
      enqueue_source_txt(st->msg);
      }
   }

char *owt_file_header(void) {
   char *msg = malloc(MXSTRMTXTLEN);
   fdf_hdr_string(msg);
   return msg;
   }

void update_interface_stats(uint32_t now_ts) {
   /* Called every STATS_INTERVAL */
   int pd = pkt_drops();
   struct source_txt *st = get_source_txt_object(&source_txt_space);

   if (st != NULL) {
      int j, av_pkt_count, n_seconds, mx_pkt_rate;
      struct if_info *iip;

      av_pkt_count = n_seconds = mx_pkt_rate = 0;
      mutex_lock("uis", &if_info_mutex);
      for (j = 0; j != n_interfaces; ++j) {
         iip = &if_info[j];
         av_pkt_count += iip->av_pkt_count;
         if (iip->n_seconds > n_seconds) n_seconds = iip->n_seconds;
         if (iip->mx_pkt_rate > mx_pkt_rate) mx_pkt_rate = iip->mx_pkt_rate;
         iip->mx_pkt_rate = iip->av_pkt_count = iip->n_seconds = 0;
         }
      mutex_unlock("uis", &if_info_mutex);

      double apc = av_pkt_count == 0 ? 1.0 : av_pkt_count;
      double r_hc_searches = n_hc_searches == 0 ? 1.0 : (double)n_hc_searches;
      double acl = (double)av_cl_count/r_hc_searches;
      uint32_t aps = (n_seconds == 0) ? av_pkt_count
         : (av_pkt_count*10+5)/(n_seconds*10);
      sprintf(st->msg, "#Stats: %u  aps=%u mps=%u lpk=%u lpp=%.3f" \
            "  mpi=%u msi=%u  acl=%.3f mso=%u fi4=%.1f fi6=%.2f" \
	    "  nip=%u nfr=%u nts=%u  mql=%u",
         now_ts,  aps, mx_pkt_rate, pd, (double)pd*100.0/apc,
         mx_active_pkt_infos, mx_active_source_infos,
            acl, mx_active_sources,
	    (double)mx_active_v4_flow_infos*100.0/mx_active_source_infos,
	    (double)mx_active_v6_flow_infos*100.0/mx_active_source_infos,
         n_not_ip, n_frags, n_too_small, mx_active_opq_entries);
      enqueue_source_txt(st->msg);
      }
   n_not_ip = n_frags = n_too_small = n_no_local_address =
      av_cl_count = n_hc_searches = n_v6_opaque =
      mx_active_pkt_infos = mx_active_source_infos =
      mx_active_sources = mx_active_v4_flow_infos = mx_active_v6_flow_infos =
      mx_active_opq_entries = 0;

#if MKSTATSCOUNTS  /* Move on to next interval */
   active_sources_this_interval = 
      active_sources_per_stats_interval[stats_interval_nbr];
   unique_hosts_this_interval =
      unique_hosts_per_stats_interval[stats_interval_nbr];
   //x   pkts_seen_this_interval = 
   //x      pkts_seen_per_stats_interval[stats_interval_nbr];
   ++stats_interval_nbr;
#endif

#if MONITOR_PKTD_STATE
   gather_source_pkt_stats();
#endif
   }

static void each_time_rec(uint32_t now_ts) {
   /* Called every TIME_INTERVAL */
   output_time_record(now_ts);

   struct flow_watcher_control *fw;  int j;
   struct source_txt *st;
   for (j = 0; j != n_fw_running; ++j) {
      fw = &flow_watcher_control_space.objects[j];
      if ((st = get_source_txt_object(&source_txt_space)) == NULL)
         log_msg(LOG_ERR, 1, "each_time_rec() couldn't get source_txt");
      count_source_states(st->msg, fw, "#SrcStates:"); 
      enqueue_source_txt(st->msg);
      }

#if SOURCETIMEOUTS  /* For trace file, just write summary at EOF */
   if (now_ts - start_ts < 10) log_msg(LOG_WARNING, 0,
      "Trace starts < 10 s before hour, not writing sum2 file yet");
   else {
      if ((st = get_source_txt_object(&source_txt_space)) == NULL)
         log_msg(LOG_ERR, 1, "each_time_rec() couldn't get source_txt");
      sprintf(st->msg, "#Record: time to write a sum file!");
      enqueue_source_txt(st->msg);
      }
#endif
   }

int fw_id = 0;
void start_flow_watcher(struct flow_watcher_control *fw, 
      uint32_t pw_q_sz, int ht_size) {
   uint32_t j;
   fw->id = fw_id;  fw_id += 1;
#if STARTUP_DEBUG
      printf("start_flow_watcher(1): pw_q_sz=%d, ht_size=%d\n",  //dd
   	  pw_q_sz, ht_size);  //dd
#endif
   fw->sh_table = owt_ht_create(ht_size);  /* Source hash table */
   if (fw->sh_table == NULL)
      log_msg(LOG_ERR, 1, "Couldn't create hash table <<<");

   fw->ows_count = 0;  // XXX
   init_pkt_info_queue(&fw->ippq);
   init_source_queue(&fw->sq);
   init_source_queue(&fw->sq2);
   t_semaphore_init(&fw->ippq_ready);
   t_semaphore_decrement(&fw->ippq_ready);  /* Thread blocked (v=0) */
   t_semaphore_init(&fw->ippq_empty);
   for (j = 0; j != pw_q_sz-1; ++j)
      t_semaphore_up(&fw->ippq_empty);  /* v=(nbr entries avail) */

   mutex_lock("sfw", &n_fw_mutex);
   n_fw_running += 1;
   mutex_unlock("sfw", &n_fw_mutex);

   if ((fw->eof_pp = get_pkt_info_object(&pkt_info_space)) == NULL)
      log_msg(LOG_ERR, 1, "Couldn't get source in start_flow_watcvher()\n");
#if STARTUP_DEBUG
   printf("start_fw() about to call pthread_create()\n"); fflush(stdout);
#endif
   if (pthread_create(&fw->flow_thread, NULL, flow_watcher_process, fw)) {
      fprintf(stderr, "Error creating flow-watch thread <<<\n");
      exit(41);
      }
#if STARTUP_DEBUG
   printf("leaving start_fw +-+\n"); fflush(stdout);
#endif
   }

int n_flow_watchers, fw_next_st_to_check;

FILE *tcp_keys_f;  //tcpk

void start_meter(char *interfaces[], int n_ifs,
   uint32_t mx_events, uint32_t mx_opq_entries, uint32_t mx_sources,
   uint32_t pw_q_sz, uint32_t mx_source_txts, uint32_t mx_flow_watchers,
   uint32_t mx_source_types, uint32_t hours_reqd)
{
   struct event *ep;
   char name[250], *np;
   int mx_pkt_infos, mx_source_infos, mx_flow_infos, err, r, j;
   uint32_t ht_size, k;
   struct if_info *iip;

   //   tcp_keys_f = fopen("tcp_keys.txt", "w");  //tcpk

#if STARTUP_DEBUG
   printf("entering start_meter() +-+\n");  fflush(stdout);
#endif
   sprintf(name, "One-Way Traffic Monitor v%s", OWTversion);
   n_source_types = mx_source_types;

#if MKSTATSCOUNTS
#define N_INTERVALS  (traces_to_process*STATS_INTERVALS_PER_HOUR +1)
   active_sources_per_stats_interval =  /* Allocate per-src_type rows */
      calloc(N_INTERVALS, sizeof(uint32_t *));
      /* +1 because we see a few packets recorded after the hour change! */
   unique_hosts_per_stats_interval =
      calloc(N_INTERVALS,  sizeof(uint32_t *));
   for (j = 0; j != N_INTERVALS; ++j) {
      /* Allocate per-minute rows */
      active_sources_per_stats_interval[j] =
	calloc(n_source_types, sizeof(uint32_t));
      unique_hosts_per_stats_interval[j] =
	calloc(n_source_types, sizeof(uint32_t *));
      }
   active_sources_this_interval = active_sources_per_stats_interval[0];
   unique_hosts_this_interval = unique_hosts_per_stats_interval[0];
   //x   pkts_seen_this_interval = pkts_seen_per_stats_interval[0];
   stats_interval_nbr = 1;  /* 1-org interval nbrs */
#endif
   n_unanalysed_sources = n_unanalysed_pkts = 0;
#if MKDESTCOUNTS
   dest_addrs_per_source_type = malloc(n_source_types*sizeof(uint32_t *));
   for (j = 0; j != n_source_types; ++j) {
      dest_addrs_per_source_type[j] = calloc(NDESTBINS, sizeof(uint32_t));
      }
#endif

   request_shutdown = ct_count = 0;

   mx_pkt_infos = 5000;
#if SOURCETIMEOUTS
   mx_source_infos = mx_sources; /* was *8/10;  17 Jan 12 */
   if (mx_source_infos < 30000) mx_source_infos = 300000;
   mx_flow_infos = mx_source_infos/3;
#else
   mx_source_infos = mx_sources;
   mx_flow_infos = mx_sources;
#endif
   owt_start(mx_events, mx_opq_entries, mx_pkt_infos,
      mx_sources, mx_source_infos, mx_flow_infos,
      mx_source_txts, mx_flow_watchers);
   last_timerec_sec = 0;  uname(&node);

   mx_active_sources = mx_active_v4_flow_infos = mx_active_v6_flow_infos =
      mx_active_opq_entries = 0;
   n_interfaces = n_ifs;
   for (err = j = 0; j != n_interfaces; ++j) {
      iip = &if_info[j];
      iip->if_name = interfaces[j];
      if ((r = init_trace(iip)) != 0)  /* Do this before scheduling events! */
	 err = 1;
      iip->if_nbr = j;  iip->n_pkts = 0;
      iip->mx_pkt_rate = iip->av_pkt_count = iip->n_seconds = 0;
      }
   if (err) exit(99);
   n_not_ip = n_frags = n_too_small = n_v6_opaque =
      intervals_this_sum_file = 0;

   sprintf(name, "%s", interfaces[0]);
   np = name + strlen(name);
   for (j = 1; j < n_ifs; ++j) {
      sprintf(np, ", %s", interfaces[j]);
      np = name + strlen(name);
      }
   sprintf(np, " on %s", node.nodename);
   trace_name = malloc(strlen(name)+1);
   strcpy(trace_name, name);

   t_semaphore_init(&opq_ready);
   t_semaphore_decrement(&opq_ready);  /* Thread blocked (v=0) */
   t_semaphore_init(&opq_empty);
   for (k = 0; k != opq_entry_space.mx_objects-1; ++k)
       t_semaphore_up(&opq_empty);  /* v=(nbr entries avail) */

   output_fdf_hdr();  /* Output .fdf header record(s) */

   ep = get_event(S_TIME_EVENT, 1);    /* Every 1s */ 
   schedule_event(ep, start_ts+3);  /* First time 3s from now */

   ep = get_event(STATS_EVENT, STATS_RECORDING_INTERVAL);
   schedule_event(ep, start_ts+4);

   ep = get_event(TIME_EVENT, time_rec_interval);
   schedule_event(ep, (start_ts+time_rec_interval-1)
      /time_rec_interval*time_rec_interval);  // 12 Apr 10, was +1

   n_flow_watchers = flow_watcher_control_space.mx_objects;
   n_fw_running = 0;
   fw_next_st_to_check = n_flow_watchers;
   reset_summary_info();
   last_drops = 0;

#if STARTUP_DEBUG
   printf("about to start flow-watch threads +-+\n");  fflush(stdout);
#endif
   /* Start flow-watching thread(s) */
   ht_size = source_space.mx_objects*(MX_EXTRA_BLOCKS+1)*2/mx_flow_watchers;
   if (ht_size < MIN_HT_SIZE) ht_size = MIN_HT_SIZE;
#if STARTUP_DEBUG
      printf("start_meter(2.5):mx_source=%u, mx_fw=%u, ht_size=%u\n",
         source_space.mx_objects,mx_flow_watchers, ht_size);  //d
#endif
   for (k = 0; k != flow_watcher_control_space.mx_objects; ++k)
      start_flow_watcher(&flow_watcher_control_space.objects[k],
         pw_q_sz, ht_size);

#if STARTUP_DEBUG
   printf("about to start packet-wat threads +-+\n");  fflush(stdout);
#endif
   /* Start the packet-watching thread(s) */
   for (n_interfaces_running = j = 0; j != n_interfaces; ++j) {
      iip = &if_info[j];
      if (pthread_create(&iip->pw_thread, NULL, packet_watcher_process, iip)) {
         fprintf(stderr, "Error creating packet-watch thread <<<\n");
         exit(31);
         }
      iip->thread_running = 1;  n_interfaces_running += 1;
      }

   setgid(ut_gid);  setuid(ut_uid);  /* Don't need to run as root now */
   my_uid = ut_uid;  my_gid = ut_gid;

   log_msg(LOG_WARNING, 0, "Starting %s, running on %s",
      OWTversion, node.nodename);

#if STARTUP_DEBUG
   printf("leaving start_meter() +-+\n");  fflush(stdout);
#endif
   }

void stop_meter(char const *msg) {
   struct source_txt *st;
   st = get_source_txt_object(&source_txt_space);
   strcpy(st->msg, msg);
   enqueue_source_txt(st->msg);  /* Queue shutdown message */
   }

int check_source_chains(char const *msg, struct source_queue *sq)
{  /* Consistency check for source queue */
   struct source *fsp,*rsp, *l_fsp, *l_rsp;
   int k, nas, rv;

   nas = sq->length;
   fsp = sq->head;  rsp = sq->tail;
   l_fsp = l_rsp = NULL;

   for (rv = 1, k = 0; ; ++k) {  /* Walk forward and reverse chains */
      if (fsp == NULL || rsp == NULL) {  /* Both NULL when done */
         if (fsp != NULL || fsp != rsp) {
	    log_msg(LOG_WARNING, 0,
               "s_c_c(%s): fsp=%x != rsp=%x, k=%d. nas=%d\n",
		    msg, fsp, rsp, k, nas);
 quack(419);
            rv = 0;
            }
         break;
         }
      if (k > nas+10) {
         log_msg(LOG_WARNING, 0,
            "s_c_c(%s): loop in chain, k=%d, nas=%d", msg, k, nas);
         fflush(stdout);
 quack(420);
         rv = 0;
         break;
         }
      l_fsp = fsp;  l_rsp = rsp;  /* For gdb debugging */
      fsp = fsp->next;  rsp = rsp->prev;
      }
   printf("s_c_C *** k=%d, nas=%d\n", k, nas);
   if (k != nas) {
      log_msg(LOG_WARNING, 0,
         "s_c_c(%s): k=%u != nas=%u", msg, k, nas);
      printf("s_c_c(%s): k=%u != nas=%u", msg, k, nas);
 quack(421);
      printf("\n");
      exit(1234);
      rv = 0;
      }
   return rv;
   }

double at_seconds(struct timeval tv) {  /* Arrival time in seconds (us precision) */
   return (double)(tv.tv_sec-start_ts) + (double)tv.tv_usec/1000000.0;
      /* Arrival times are relative to start_ts */
   }

uint32_t pkt_drops(void) {
   int j;  uint64_t r, drops = 0;
   for (j = 0; j != n_interfaces; ++j)
      drops += trace_get_dropped_packets(if_info[j].trace);
   r = drops < last_drops ? 0 : drops - last_drops;
   last_drops = drops;
   return (uint32_t)r;
   }

int init_trace(struct if_info *iip) {
  struct timeval tv;  uint32_t ts;  int fl;

   iip->trace = NULL;  iip->packet = NULL;
   iip->trace = trace_create(iip->if_name);
   if (trace_is_err(iip->trace)) {
      log_msg(LOG_ERR, 0, "Couldn't open trace %s <<<", iip->if_name);
      return 101;
      }

   snaplen = SNAP_LEN;
   if (trace_config(iip->trace,TRACE_OPTION_SNAPLEN,&snaplen)) {
      log_msg(LOG_ERR, 0, "Couldn't set snap length %d <<<", snaplen);
      return 102;
      }

   if (trace_start(iip->trace)) {
      log_msg(LOG_ERR, 0, "Couldn't start trace %s <<<", iip->if_name);
      trace_destroy(iip->trace);
      return 103;
      }

   iip->packet = trace_create_packet();
   for (;;) {  /* Some sources take a while to get valid timestamps! */
      if (trace_read_packet(iip->trace, iip->packet) > 0) {
	 fl = (int)trace_get_framing_length(iip->packet);
         if (fl != 0) {  /* Ignore rt: startup packets */
            tv = trace_get_timeval(iip->packet);
            start_ts = ts = (uint32_t)tv.tv_sec;
            set_date_seed(start_ts);  /* For Conficker p2p test */
            break;
	    }
         }
      else {
         log_msg(LOG_ERR, 0, "Couldn't read packet from %s <<<", iip->if_name);
         return 104;
         }
      }   
   return 0;
   }

void close_trace(struct if_info *iip) {
   if (iip->trace != NULL) {
      if (trace_is_err(iip->trace)) log_msg(LOG_WARNING, 0,
         "Close_trace(): found error  %s !!!", iip->if_name);
      trace_destroy(iip->trace);
      }
   if (iip->packet != NULL)
      trace_destroy_packet(iip->packet);
   }

int next_trace_file(struct if_info *iip) {
   free(iip->if_name);  /* processewd uri */
   iip->if_name = other_uris[next_uri_nbr];
   next_uri_nbr += 1;

   if (init_trace(iip) != 0)
      log_msg(LOG_ERR, 1, "Unable to start trace %s <<<", iip->if_name);

   return 0;
   }

void one_second_process(uint32_t ts) {
   struct if_info *iip;
   uint32_t active_opq_entries, active_pkt_infos,
      active_sources, active_source_infos,
      active_v4_flow_infos, active_v6_flow_infos;
   int j;

   last_ts = ts;  /* Save ts for end-of-run */
   set_date_seed(ts);  /* For Conficker p2p test */

   for (j = 0; j != n_interfaces; ++j) {
      iip = &if_info[j];
      iip->av_pkt_count += iip->n_pkts;  iip->n_seconds += 1;
      if (iip->n_pkts > iip->mx_pkt_rate) iip->mx_pkt_rate = iip->n_pkts;
      iip->n_pkts = 0;  g_last_ts = iip->last_ts = ts;
      }
   mutex_unlock("osp", &if_info_mutex);  /* Locked in seconds_change() */

   active_opq_entries = opq_entry_space.mx_objects -
      opq_entry_space.free_queue.length;
   if (active_opq_entries > mx_active_opq_entries)
      mx_active_opq_entries = active_opq_entries;

   active_pkt_infos = pkt_info_space.mx_objects -
      pkt_info_space.free_queue.length;
   if (active_pkt_infos > mx_active_pkt_infos)
      mx_active_pkt_infos = active_pkt_infos;

   active_sources = source_space.mx_objects -
      source_space.free_queue.length;
   if (active_sources > mx_active_sources)
      mx_active_sources = active_sources;
   active_source_infos = source_info_space.mx_objects -
      source_info_space.free_queue.length;
   if (active_source_infos > mx_active_source_infos)
      mx_active_source_infos = active_source_infos;

   active_v4_flow_infos = v4_flow_info_space.mx_objects -
      v4_flow_info_space.free_queue.length;
   if (active_v4_flow_infos > mx_active_v4_flow_infos)
      mx_active_v4_flow_infos = active_v4_flow_infos;
   active_v6_flow_infos = v6_flow_info_space.mx_objects -
      v6_flow_info_space.free_queue.length;
   if (active_v6_flow_infos > mx_active_v6_flow_infos)
      mx_active_v6_flow_infos = active_v6_flow_infos;
   }

#if STOP_AFTER
int n_times = STOP_AFTER;
#endif
void seconds_change(struct if_info *iip, uint32_t ts) {
   uint32_t e_type;
   struct pkt_info *event_pp;
   struct flow_watcher_control *fw;

   mutex_lock("osp", &if_info_mutex);
   if (ts <= g_last_ts) {
      iip->last_ts = ts;
      mutex_unlock("osp", &if_info_mutex);
      }
   else {  /* Change in overall last_ts */
      one_second_process(ts);  /* Update g_last_ts, unlock if_info */

      for (;;) {  /* Handle events */
         e_type = check_events(ts);
         if (e_type == NO_EVENT) break;
	 else switch (e_type) {
	 case S_TIME_EVENT:  /* Timeout sources */
	    if ((event_pp = get_pkt_info_object(&pkt_info_space)) == NULL)
	       break;  /* Can't schedule another TIME_EVENT */

   	    fw_next_st_to_check -= 1;
	    fw = &flow_watcher_control_space.objects[
               fw_next_st_to_check];
            if (fw_next_st_to_check == 0)
               fw_next_st_to_check = n_flow_watchers;

	    t_semaphore_down(&fw->ippq_empty);
            event_pp->event_type = S_TIME_EVENT;
            event_pp->ts = g_last_ts;

            pkt_info_q_lock(&fw->ippq);
            pkt_info_push(&fw->ippq, event_pp);
            pkt_info_q_unlock(&fw->ippq);
            t_semaphore_up(&fw->ippq_ready);
               /* Pass event to flow watcher */
            break;
	 case STATS_EVENT:  /* Write #Stats */
	    update_interface_stats(ts);
#if STOP_AFTER
            if (--n_times == 0) request_shutdown = 1;
#endif
            break;
         case TIME_EVENT:  /* Write #Time */
	    each_time_rec(ts);
            break;
            }
         }  /* event handling loop */
      }
   }

int n_tcpks = 0;    //tcpk

void each_pkt(struct flow_watcher_control *fw, struct pkt_info *ipp) {
   struct source *slr, *tq;
   int new = 1, already_moved = 0;
   uint32_t hash;

   hash = owt_hash(ipp);
   slr = owt_ht_lookup(fw->sh_table, hash, ipp, &new);
                                         /* new => OK_to_insert */
   if (new) {  /* New source, slr == isp */
     // printf("+");  fflush(stdout);  // WWW
      slr->state = SS_ONE_WAY;
      slr->fw_nbr = fw->id;
      slr->d.si->timeout = STO_ONE_WAY;
      slr->version = ipp->version;
      slr->first_fwd = ipp->fwd;
      slr->d.si->tb = slr->d.si->fb = slr->d.si->tp =
         slr->d.si->fp = slr->totp = 0;
      slr->d.si->first_s = ipp->at_s;
      slr->d.si->last_report_sec = ipp->ts;
      slr->src_addr = ipp->src_addr;
      slr->d.si->inactive_last_sum = 0;
      slr->d.si->tcp_key = ipp->tcp_key;  /* Only want the SYN key */
#if ERF_DIRECTION
      slr->d.si->direction = ipp->direction;
#endif
      get_flow_info(slr);
      //      if (slr->proto == 6) fprintf(tcp_keys_f, "%8x\n", slr->tcp_key);  //tcpk
      //      if ((n_tcpks += 1) == 10000) { fclose(tcp_keys_f);  exit(123); }  //tcpk
      }
   else {  /* Already in table, copy fields from ipp */
      if (slr->state == SS_WAS_TWO_WAY) {
 	 source_q_lock(&fw->sq2);  /* Move source back into 'normal' queue */
	 tq = source_extract(&fw->sq2, slr);
	    source_q_unlock(&fw->sq2);
	 source_q_lock(&fw->sq);
         source_push(&fw->sq, tq);
	    source_q_unlock(&fw->sq);
	 already_moved = 1;
 	 if (slr->d.si == NULL)
            slr->d.si = get_source_info();
         else {
#if OWS_STATS
            slr->d.si->otp = slr->d.si->tp;  slr->d.si->ofp = slr->d.si->fp;
            slr->d.si->ototp = slr->totp;
            slr->d.si->orate = slr->totp == 1 ? 0 : 
               slr->totp/(slr->d.si->last_s - slr->d.si->first_s);
#endif
            slr->d.si->first_s = slr->d.si->last_s;
               /* Time we saw last pkt before timeout */
	    }
	 slr->state = SS_NEW_ONE_WAY;
         fw->ows_count += 1;
         slr->version = ipp->version;
         slr->first_fwd = ipp->fwd;
         slr->d.si->tb = slr->d.si->fb = slr->d.si->tp =
            slr->d.si->fp = slr->totp = 0;
         slr->d.si->last_report_sec = ipp->ts;
         slr->d.si->timeout = STO_ONE_WAY;
         slr->src_addr = ipp->src_addr;
         slr->d.si->inactive_last_sum = 0;
         slr->d.si->tcp_key = ipp->tcp_key;  /* Only want the SYN key */
#  if ERF_DIRECTION
         slr->d.si->direction = ipp->direction;
#  endif
         get_flow_info(slr);
         }

#if ERF_DIRECTION  /* Keep track of direction for *all* pkts */
      slr->d.si->direction |= ipp->direction;
#endif
      }
   slr->ts = ipp->ts;
   slr->d.si->last_s = ipp->at_s;

   if (ipp->fwd) {  /* !local -> local, i.e. pkts in to home network */
      mutex_lock("epk", &source_data_mutex);
      slr->d.si->tp += 1;  slr->d.si->tb += ipp->ip_len;
      mutex_unlock("epk", &source_data_mutex);
      }
   else {
      mutex_lock("epk", &source_data_mutex);
      slr->d.si->fp += 1;  slr->d.si->fb += ipp->ip_len;
      mutex_unlock("epk", &source_data_mutex);
      }
   slr->totp += 1;
   slr->d.si->active_since_last_sum = 1;

   if (slr->state == SS_ONE_WAY || slr->state == SS_NEW_ONE_WAY) {  /* Not yet SS_TWO_WAY */
      if (slr->d.si->fp != 0) {  /* Pkts out from home network -> two-way source */
         fw->ows_count -= 1;  // XXX
         slr->state = SS_TWO_WAY;
         slr->d.si->timeout = STO_TWO_WAY;
 	 free_flow_info(slr);  /* Don't need flow info now */
         }
      else {  /* Still one-way, update flow info */ 
	 /* SRI test for Conficker p2p source ... */ 
	 ipp->cfc = 0;
	 if (slr->d.si->flow_info.fi4) {  /* i.e. flow_info != NULL */ 
	    /* Any UDP, any TCP not to port 445 or from port 80 */ 
	    if (ipp->version == 4 && (
	          ipp->proto == PT_UDP || (ipp->proto == PT_TCP && 
		  !(ipp->dst_port == 445 || ipp->src_port == 80)) )) {
	       int rc, i;  result_t res; 
	       rc = portgen(ipp->dst_addr.a.v4, &res, date_seed); 
	       for (i = 0; i < 8; i++) { 
	          if (res.u16[i] == ipp->dst_port) { 
		     ipp->cfc = 1;  /* For updating flow_info */
		     break;  /* Need only match one port! */ 
	             } 
	          } 
	       } 
	    }
         int half_flow = update_flow_info(slr, ipp);
	 
#if HALF_WAY_SOURCES 
         if (slr->state == SS_ONE_WAY && half_flow) { 
	    fw->ows_count -= 1;  // XXX 
	    slr->state = SS_HALF_WAY; 
	    slr->d.si->timeout = STO_ONE_WAY; 
	    free_flow_info(slr);  /* Don't need flow info now */ 
	    } 
#endif   
         } 
      } 
#if HALF_WAY_SOURCES
   else if (slr->state == SS_HALF_WAY) {
      if (slr->d.si->fp != 0) {  /* Pkt out from home -> two-way source */
	 // BBB         fw->ows_count -= 1;  // XXX
         slr->state = SS_TWO_WAY;
         slr->d.si->timeout = STO_TWO_WAY;
         }
      }
#endif
   else if (slr->state == SS_DELETE) {
      }
   else if (slr->state == SS_WAS_TWO_WAY) {
      }
   else if (slr->state == SS_NEW_ONE_WAY) {
      }
   else if (slr->state == SS_TWO_WAY) {
      }
   else log_msg(LOG_ERR, 0, "source in state %d <<<<", slr->state);

   free_pkt_info(ipp);  /* Finished with ipp */
   if (new) {
      source_q_lock(&fw->sq);
      source_push(&fw->sq, slr);
      source_q_unlock(&fw->sq);
      }
   else {
      if (!already_moved) {
         source_q_lock(&fw->sq);
         tq = source_extract(&fw->sq, slr);  /* Move slr to source list tail */
            /* check_source_timeouts() starts from head,
	       so walk_up() starts from the most recently active */
         source_push(&fw->sq, tq);
         source_q_unlock(&fw->sq);
         }
      }
#if SC_DEBUG >= 3
   check_source_chains("+++ end of each_pkt", &fw->sq);
#endif
   }

void shutdown_flow_watchers(void) {
   struct flow_watcher_control *fw;
   int k;
   for (k = 0; k != n_flow_watchers; ++k) {
      // printf("shutdown_flow_watchers: k=%d\n", k);
      fw = &flow_watcher_control_space.objects[k];
      t_semaphore_down(&fw->ippq_empty);  /* Wait for an empty entry */
      fw->eof_pp->event_type = END_OF_FILE;
      pkt_info_q_lock(&fw->ippq);
      pkt_info_push(&fw->ippq, fw->eof_pp);
      pkt_info_q_unlock(&fw->ippq);
      t_semaphore_up(&fw->ippq_ready);  /* Signal a full entry */
      }
   }

void cancel_packet_watchers(void) {
   int j, r;
   for (j = 0; j != n_interfaces; ++j) {
      if (if_info[j].thread_running) {
         r = pthread_cancel(if_info[j].pw_thread);
         printf("thread %d cancel -> %d\n", j,r);  fflush(stdout);
         }
      }
   shutdown_flow_watchers();
   }

uint32_t nfw[4] = {0, 0, 0, 0};

void *packet_watcher_process(void *arg) {
   struct if_info *iip = (struct if_info *)arg;
   struct timeval tv;
   uint16_t ethertype;   uint32_t hl;  libtrace_ip_t *ip;
   int r, key_code, fwd, fw_nbr;
   struct pkt_info *new_pp;
   uint32_t ts;
   struct flow_watcher_control *fw;
   int fn = 0, restarts = 0;

   //##   fwd = nice(-5);  /* Run this thread at high priority */
      /* Result not used, keep compiler happy */

   iip->last_ts = start_ts;
   for (;;) {  /* Loop through files in list */
      for (;;) {  /* int_trace() leaves first packet in iip->packet */
         if (request_shutdown)  /* User sent SIGINT */
       	    break;  /* Test here; pthread_cancel seems to clobber iip->trace */
         iip->n_pkts += 1;  /* For av/max pkts #stats */
         ip = trace_get_layer3(iip->packet, &ethertype, &hl);
         if (!ip || (ethertype != 0x0800 && ethertype != 0x86dd))
	    n_not_ip += 1;
         else {
            tv = trace_get_timeval(iip->packet);
            ts = (uint32_t)tv.tv_sec;
	    if (ts > iip->last_ts) seconds_change(iip, ts);
            key_code = ethertype == 0x0800 ?
	       get_v4key(&new_pp, &fwd,  hl, (uint8_t *)ip)
 	       : get_v6key(&new_pp, &fwd,  hl, (uint8_t *)ip);
            if (key_code != SKEY_OK) {
               /* Couldn't get packet key (packet freed by get_key) */
               switch (key_code) {
               case SKEY_NOPEERADDR:
               case SKEY_NOTRANSADDR:
	          n_too_small += 1;
                  break;
               case SKEY_FRAGPKT:
                  n_frags += 1;
                  break;
               case SKEY_NOLOCALADDR:
                  n_no_local_address += 1;
                  break;
               case SKEY_V6OPAQUE:
                  n_v6_opaque += 1;
                  break;
	       case SKEY_NOSPACE:
	 	  break;  /* Ignore, get_source displayed gthe message */
               default:
 		  log_msg(LOG_ERR, 0,  /* IP version wrong for ethertype */
	             "packet_watcher: bad key_code %d !!!", key_code);
                  break;
                  }
	       }
            else {
	       fw_nbr = (n_flow_watchers == 1 ? 0 :
                  (new_pp->version == 4 ?
		  new_pp->src_addr.a.a[IP4_ADDR_LEN-2] :
		  new_pp->src_addr.a.a[IP6_ADDR_LEN-2]) )
		  % n_flow_watchers;
	       nfw[fw_nbr] += 1;
               fw = &flow_watcher_control_space.objects[fw_nbr];

               t_semaphore_down(&fw->ippq_empty);  /* Wait for empty entry */
               new_pp->event_type = NO_EVENT;
               new_pp->fwd = fwd;
               new_pp->ts = ts - start_ts;
               new_pp->wl = trace_get_wire_length(iip->packet);
               new_pp->at_s = at_seconds(tv);
#if ERF_DIRECTION
 	       new_pp->direction = trace_get_direction(iip->packet)+1;
	          /* 0 => portA, 1 => portB */
#endif
               pkt_info_q_lock(&fw->ippq);
               pkt_info_push(&fw->ippq, new_pp);
               pkt_info_q_unlock(&fw->ippq);
               t_semaphore_up(&fw->ippq_ready);
                  /* Pass pkt_info to selected flow watcher */
   	       }
            }

          r = trace_read_packet(iip->trace, iip->packet);  /* Next packet */
         if (r <= 0) {
	    if (r == 0) break;  /* EOF */
            log_msg(LOG_WARNING, 0, "Error %d from read_packet()");

            if (restarts == 3)
               log_msg(LOG_WARNING, 1, "3 restart attempts, stopping meter!");
            else {
               close_trace(iip);
	       if (init_trace(iip) != 0)
 		  log_msg(LOG_WARNING, 1, "Couldn't restart trace, quitting!");
               else {
		  r = trace_read_packet(iip->trace, iip->packet);
                  if (r > 0) restarts += 1;  /* OK now */
                  else log_msg(LOG_WARNING, 1,
                     "Couldn't read packet after restart");
	          }
	       }

	    }
         }  /* SIGINT or trace file EOF */

      if (request_shutdown) break;  /* User sent SIGINT */
      fn += 1;
      if (fn == traces_to_process) break;  /* No more files */

      close_trace(iip);
      if (next_trace_file(iip) != 0)
         log_msg(LOG_ERR, 1, "Couldn't start next file !!!");
      log_msg(LOG_WARNING, 0, "----- opened %s", iip->if_name);
      }

   iip->thread_running = 0;
   n_interfaces_running -= 1;
   // printf("EOF, n_interfaces_ruunning=%d\n", n_interfaces_running);
   if (n_interfaces_running == 0)
      shutdown_flow_watchers();
   return NULL;  /* Keep compiler happy */
   }

void *flow_watcher_process(void *arg) {
   struct flow_watcher_control *fw = arg;
   struct pkt_info *ipp;
   int shutdown_now, rake_reqd, j;

   //##   j = nice(-3);  /* Run this thread at medium priority */
      /* Result not used, keep compiler happy */
#if STARTUP_DEBUG
   printf("starting flow_watcher_process() +-+\n"); fflush(stdout); 
#endif
   for (;;) {
      t_semaphore_down(&fw->ippq_ready);  /* Wait for a full block */
      pkt_info_q_lock(&fw->ippq);
      ipp = pkt_info_shift(&fw->ippq);
      pkt_info_q_unlock(&fw->ippq);

      if (ipp->event_type == END_OF_FILE) {
 	 // printf("flow_watcher(%d): EOF\n", fw->id);
         break;
         }
      else switch (ipp->event_type) {
      case S_TIME_EVENT:
        rake_reqd = n_extra_sources == MX_EXTRA_BLOCKS &&  
	   source_space.free_queue.length < MN_FREE_SOURCES;
	if (SOURCETIMEOUTS || rake_reqd) {
            mutex_lock("recording", &count_mutex);
	    if (CV_RECORDING)  /* Handling #Record: don't try to rake */
	       mutex_unlock("recording", &count_mutex);
	    else {
  	       ct_count += 1;  /* About to run check_source_timeouts() */
               mutex_unlock("recording", &count_mutex);
	       if (rake_reqd)
                  j = rake_idle_sources(fw, ipp->ts);
               else if (SOURCETIMEOUTS) {
                  check_source_timeouts(fw, ipp->ts);
                  check_was_two_way_timeouts(fw, ipp->ts);
	          }

    	       pthread_mutex_lock(&count_mutex);
	       ct_count -= 1;  /* check_source_timeouts() finished */
 	       pthread_cond_signal(&count_cv);  /* Let Record: know! */
	       pthread_mutex_unlock(&count_mutex);
	       }
	    }
	 ipp->event_type = NO_EVENT;
	 free_pkt_info(ipp);
	 break;
      default:
	 each_pkt(fw, ipp);
	 break;
         }

      t_semaphore_up(&fw->ippq_empty);  /* Signal an empty block */
      }  /* Loop finishes on interrupt or end of trace file */

      //xx  output_active_sources(&fw->sq, isp->ts);
      /* Dump currently active sources */

   mutex_lock("fwp", &n_fw_mutex);
   n_fw_running -= 1;
   shutdown_now = n_fw_running == 0;
   mutex_unlock("fwp", &n_fw_mutex); 

   if (shutdown_now) {
      printf("$$$ shutting down ...\n");  // XXX
      struct flow_watcher_control *fw;
      int t_one_way = 0;  char buf[120];
      for (j = 0; j != n_flow_watchers; ++j) {
         fw = &flow_watcher_control_space.objects[j];
         count_source_states(buf, fw, "#SrcStates:"); 
         printf("%s\n", buf);
         }
      for (j = 0; j != n_interfaces; ++j)
         close_trace(&if_info[j]);
      if (request_shutdown) 
         stop_meter("#Shutdown: User sent interrupt ###");
      else {
	 printf("nfw[] = %u, %u, %u, %u\n", nfw[0], nfw[1], nfw[2], nfw[3]);
         stop_meter("#Shutdown: End of trace file ###");
         }
      }
   return NULL;  /* Keep compiler happy */
   }
