#!/usr/bin/env ruby

# 1620, Tue 28 Jun 11 (NZST)
#
#  src-list-split.rb: Make separate .bpf files for list of src_addrs
#                     in a single output trace file
#
#  Copyright 2011-2012, Nevil Brownlee, U Auckland | CAIDA

require 'RubyLibtrace'
require './owt-rb/os-times'

class OutTrace
  def initialize(tr_fn, spec_fn, trac_fn)
    @src_addrs = Hash.new
    spf = File.open(spec_fn, 'r')
    spf.each do |line|
      aa = line.split
      next if aa[0] == '#'
      sa = IPprefix.from_s(aa[0])
      print "#{sa}\n"
      @src_addrs[sa.addr] = true
    end
    spf.close

    @ot = OutputTrace.new("pcapfile:#{trac_fn}.pcap.gz")
    # ot.conf_compress_type(ZLIB_COMPRESSION)
    # ot.conf_compress_level(5)
    @ot.start_output

    @pt = Trace.new(tr_fn)
    @pt.start
  end

  def process
    np = 0
    @pt.each_packet do |pkt|
      i = pkt.ip
      i = pkt.ip6  if i.nil?
      next  if i.nil?

      sa = i.src_prefix.addr;  da = i.dst_prefix.addr
      @ot.write_packet(pkt) if !@src_addrs[sa].nil? || !@src_addrs[da].nil?

      # print "np=#{np}: s=#{sa}, a=>#{a}<, #{pkt}\n"
      np += 1  #;  break if np == 20

      if np % 100000 == 0
        print '.';  $stdout.flush
      end

    end
    @ot.close_output
    @pt.close
  end
end  # OutTrace

 $st = OStimes.new

# tfn = '/data/telescope/live/2011-03-08/net44.1299596400.pcap.gz' # odin
# tfn = '/home/nevil/scratch/20110308-5m.bpf'  # nebbiolo


def run(tfn, spec_fn, sf)
  $st.reset
  ott = OutTrace.new(tfn, spec_fn, sf)
  ott.process
  $st.session_times($stdout, spec_fn)
end

tfn = 'erf:/home/jbro111/dag/test-br2-5m.erf.gz'  # nevil-res3
spec_fn = ARGV[0];  trac_fn = ARGV[1]
run(tfn, spec_fn, trac_fn)
